@@ -25,9 +25,9 @@
             for station, group in zip(STATIONS, station_groups):
                 download_data(data, group, station, START, END)
 
-            coincidences = coincidences.Coincidences(data, '/coincidences',
-                                                     station_groups)
-            coincidences.search_and_store_coincidences()
+            coin = coincidences.Coincidences(data, '/coincidences',
+                                             station_groups)
+            coin.search_and_store_coincidences()
 
 """
 import time
@@ -212,10 +212,17 @@ class Coincidences(object):
                                                  storage.EventObservables)
 
         print "Storing coincidences"
-        progress = pb.ProgressBar(widgets=[pb.Percentage(), pb.Bar(),
-                                           pb.ETA()])
-        for coincidence in progress(self.coincidence_group._src_c_index):
-            self._store_coincidence(coincidence)
+
+        # ProgressBar does not work for empty iterables.
+        if len(self.coincidence_group._src_c_index):
+            progress = pb.ProgressBar(widgets=[pb.Percentage(), pb.Bar(),
+                                      pb.ETA()])
+            for coincidence in progress(self.coincidence_group._src_c_index):
+                self._store_coincidence(coincidence)
+        else:
+            print "Creating empty tables, no coincidences found"
+            for coincidence in self.coincidence_group._src_c_index:
+                self._store_coincidence(coincidence)
 
         c_index = self.data.createVLArray(self.coincidence_group, 'c_index',
                                           tables.UInt32Col())
@@ -436,7 +443,7 @@ class CoincidencesESD(Coincidences):
 
     This subclass stores the paths to the station_groups that where
     used to look for coincidences in a lookup-table, and also
-    stores the original event_id and station info for each coincidence.
+    stores the original station info and event_id for each coincidence.
 
     """
     def search_and_store_coincidences(self, cluster=None):
@@ -482,10 +489,17 @@ class CoincidencesESD(Coincidences):
                                                   'coincidences', description)
 
         print "Storing coincidences"
-        progress = pb.ProgressBar(widgets=[pb.Percentage(), pb.Bar(),
-                                           pb.ETA()])
-        for coincidence in progress(self._src_c_index):
-            self._store_coincidence(coincidence)
+
+        # ProgressBar does not work for empty iterables.
+        if len(self._src_c_index):
+            progress = pb.ProgressBar(widgets=[pb.Percentage(), pb.Bar(),
+                                      pb.ETA()])
+            for coincidence in progress(self._src_c_index):
+                self._store_coincidence(coincidence)
+        else:
+            print "Creating empty tables, no coincidences found"
+            for coincidence in self._src_c_index:
+                self._store_coincidence(coincidence)
 
         c_index = self.data.createVLArray(self.coincidence_group, 'c_index',
                                           tables.UInt32Col(shape=2))
@@ -497,7 +511,7 @@ class CoincidencesESD(Coincidences):
         s_index = self.data.createVLArray(self.coincidence_group, 's_index',
                                           tables.VLStringAtom())
         for station_group in self.station_groups:
-            s_index.append(station_group._v_pathname)
+            s_index.append(station_group)
         s_index.flush()
 
     def _store_coincidence(self, coincidence):
